# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Register Dispatch backends"""

from .numpy import *
from .jax import *

# Set default backend if only 1 is available
from ..dispatch import Dispatch

if len(Dispatch.REGISTERED_BACKENDS) == 1:
    Dispatch.DEFAULT_BACKEND = Dispatch.REGISTERED_BACKENDS[0]

__all__ = []
